@@ -15,6 +15,7 @@ enum {
 	SIG_RESIZE_EVENT,
 	SIG_KEYBOARD_EVENT,
 	SIG_POINTER_EVENT,
+	SIG_CLIPBOARD_TEXT,
 	N_SIGS
 };
 
@@ -41,6 +42,7 @@ static void rf_vnc_server_class_init(RfVNCServerClass *klass)
 	klass->is_running = NULL;
 	klass->stop = NULL;
 	klass->set_desktop_name = NULL;
+	klass->send_clipboard_text = NULL;
 	klass->update = NULL;
 	klass->flush = NULL;
 
@@ -114,6 +116,18 @@ static void rf_vnc_server_class_init(RfVNCServerClass *klass)
 		G_TYPE_BOOLEAN,
 		G_TYPE_BOOLEAN
 	);
+	sigs[SIG_CLIPBOARD_TEXT] = g_signal_new(
+		"clipboard-text",
+		RF_TYPE_VNC_SERVER,
+		0,
+		0,
+		NULL,
+		NULL,
+		NULL,
+		G_TYPE_NONE,
+		1,
+		G_TYPE_STRING
+	);
 }
 
 static void rf_vnc_server_init(RfVNCServer *this)
@@ -172,11 +186,23 @@ void rf_vnc_server_set_desktop_name(RfVNCServer *this, const char *desktop_name)
 	klass->set_desktop_name(this, desktop_name);
 }
 
+void rf_vnc_server_send_clipboard_text(RfVNCServer *this, const char *text)
+{
+	g_return_if_fail(RF_IS_VNC_SERVER(this));
+	g_return_if_fail(text != NULL);
+
+	RfVNCServerClass *klass = RF_VNC_SERVER_GET_CLASS(this);
+	g_return_if_fail(klass->send_clipboard_text != NULL);
+
+	klass->send_clipboard_text(this, text);
+}
+
 void rf_vnc_server_update(
 	RfVNCServer *this,
 	GByteArray *buf,
 	unsigned int width,
-	unsigned int height
+	unsigned int height,
+	const struct rf_rect *damage
 )
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
@@ -186,7 +212,7 @@ void rf_vnc_server_update(
 	RfVNCServerClass *klass = RF_VNC_SERVER_GET_CLASS(this);
 	g_return_if_fail(klass->update != NULL);
 
-	klass->update(this, buf, width, height);
+	klass->update(this, buf, width, height, damage);
 }
 
 void rf_vnc_server_flush(RfVNCServer *this)
@@ -215,7 +241,7 @@ static void _iterate_keys(struct xkb_keymap *map, xkb_keycode_t key, void *data)
 		xkb_keymap_num_levels_for_key(map, key, 0);
 	for (xkb_level_index_t i = 0; i < num_levels; ++i) {
 		const xkb_keysym_t *syms;
-		int num_syms =
+		const int num_syms =
 			xkb_keymap_key_get_syms_by_level(map, key, 0, i, &syms);
 		for (int k = 0; k < num_syms; ++k) {
 			if (syms[k] == idata->keysym) {
@@ -227,6 +253,24 @@ static void _iterate_keys(struct xkb_keymap *map, xkb_keycode_t key, void *data)
 	}
 }
 
+void rf_vnc_server_handle_resize_event(
+	RfVNCServer *this,
+	unsigned int width,
+	unsigned int height
+)
+{
+	g_return_if_fail(RF_IS_VNC_SERVER(this));
+	g_return_if_fail(width > 0 && height > 0);
+
+	if (!rf_vnc_server_is_running(this))
+		return;
+
+	g_debug("VNC: Received resize event for width %d and height %d.",
+		width,
+		height);
+	g_signal_emit(this, sigs[SIG_RESIZE_EVENT], 0, width, height);
+}
+
 void rf_vnc_server_handle_keysym_event(
 	RfVNCServer *this,
 	uint32_t keysym,
@@ -235,6 +279,9 @@ void rf_vnc_server_handle_keysym_event(
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
 
+	if (!rf_vnc_server_is_running(this))
+		return;
+
 	RfVNCServerPrivate *priv = rf_vnc_server_get_instance_private(this);
 	struct _iterate_data idata = {
 		.keysym = keysym,
@@ -248,7 +295,7 @@ void rf_vnc_server_handle_keysym_event(
 		);
 		return;
 	}
-	uint32_t keycode = RF_KEY_CODE_XKB_TO_EV(idata.keycode);
+	const uint32_t keycode = RF_KEY_CODE_XKB_TO_EV(idata.keycode);
 	g_debug("Input: Received key %s for keysym %04x and keycode %u.",
 		down ? "down" : "up",
 		keysym,
@@ -264,6 +311,9 @@ void rf_vnc_server_handle_keycode_event(
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
 
+	if (!rf_vnc_server_is_running(this))
+		return;
+
 	g_debug("Input: Received key %s for keycode %u.",
 		down ? "down" : "up",
 		keycode);
@@ -284,16 +334,19 @@ void rf_vnc_server_handle_pointer_event(
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
 
-	bool left = mask & 1;
-	bool middle = mask & (1 << 1);
-	bool right = mask & (1 << 2);
-	bool wup = mask & (1 << 3);
-	bool wdown = mask & (1 << 4);
-	bool wleft = mask & (1 << 5);
-	bool wright = mask & (1 << 6);
+	if (!rf_vnc_server_is_running(this))
+		return;
+
+	const bool left = mask & 1;
+	const bool middle = mask & (1 << 1);
+	const bool right = mask & (1 << 2);
+	const bool wup = mask & (1 << 3);
+	const bool wdown = mask & (1 << 4);
+	const bool wleft = mask & (1 << 5);
+	const bool wright = mask & (1 << 6);
 	// Generally the 7th bit is reserved.
-	bool back = mask & (1 << 8);
-	bool forward = mask & (1 << 9);
+	const bool back = mask & (1 << 8);
+	const bool forward = mask & (1 << 9);
 	g_debug("Input: Received pointer at x %f and y %f, raw button %#x, "
 		"left %s, middle %s, right %s, back %s, forward %s, "
 		"wheel up %s, wheel down %s, wheel left %s, wheel right %s.",
@@ -327,19 +380,16 @@ void rf_vnc_server_handle_pointer_event(
 	);
 }
 
-void rf_vnc_server_handle_resize_event(
-	RfVNCServer *this,
-	unsigned int width,
-	unsigned int height
-)
+void rf_vnc_server_handle_clipboard_text(RfVNCServer *this, const char *text)
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
-	g_return_if_fail(width > 0 && height > 0);
+	g_return_if_fail(text != NULL);
 
-	g_debug("VNC: Emitting resize signal for width %d and height %d.",
-		width,
-		height);
-	g_signal_emit(this, sigs[SIG_RESIZE_EVENT], 0, width, height);
+	if (!rf_vnc_server_is_running(this))
+		return;
+
+	g_debug("VNC: Received clipboard text %s.", text);
+	g_signal_emit(this, sigs[SIG_CLIPBOARD_TEXT], 0, text);
 }
 
 // Those signals must be delayed to next event by using `g_idle_add()`, because
@@ -347,10 +397,10 @@ void rf_vnc_server_handle_resize_event(
 // events, it is not OK to release resources while still processing events so we
 // have to delay them until events are done.
 
-static void _emit_first_client(gpointer data)
+static void _emit_first_client(void *data)
 {
 	RfVNCServer *this = data;
-	g_debug("VNC: Emitting first client signal.");
+	g_debug("Signal: Emitting VNC first client signal.");
 	g_signal_emit(this, sigs[SIG_FIRST_CLIENT], 0);
 }
 
@@ -358,13 +408,16 @@ void rf_vnc_server_handle_first_client(RfVNCServer *this)
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
 
+	if (!rf_vnc_server_is_running(this))
+		return;
+
 	g_idle_add_once(_emit_first_client, this);
 }
 
-static void _emit_last_client(gpointer data)
+static void _emit_last_client(void *data)
 {
 	RfVNCServer *this = data;
-	g_debug("VNC: Emitting last client signal.");
+	g_debug("Signal: Emitting VNC last client signal.");
 	g_signal_emit(this, sigs[SIG_LAST_CLIENT], 0);
 }
 
@@ -372,5 +425,8 @@ void rf_vnc_server_handle_last_client(RfVNCServer *this)
 {
 	g_return_if_fail(RF_IS_VNC_SERVER(this));
 
+	if (!rf_vnc_server_is_running(this))
+		return;
+
 	g_idle_add_once(_emit_last_client, this);
 }
